@@ -35,6 +35,7 @@
 static void
 exif_mnote_data_canon_clear (ExifMnoteDataCanon *n)
 {
+	ExifMnoteData *d = (ExifMnoteData *) n;
 	unsigned int i;
 
 	if (!n) return;
@@ -42,10 +43,10 @@ exif_mnote_data_canon_clear (ExifMnoteDataCanon *n)
 	if (n->entries) {
 		for (i = 0; i < n->count; i++)
 			if (n->entries[i].data) {
-				free (n->entries[i].data);
+				exif_mem_free (d->mem, n->entries[i].data);
 				n->entries[i].data = NULL;
 			}
-		free (n->entries);
+		exif_mem_free (d->mem, n->entries);
 		n->entries = NULL;
 		n->count = 0;
 	}
@@ -59,14 +60,33 @@ exif_mnote_data_canon_free (ExifMnoteData *n)
 	exif_mnote_data_canon_clear ((ExifMnoteDataCanon *) n);
 }
 
+static void
+exif_mnote_data_canon_get_tags (ExifMnoteDataCanon *dc, unsigned int n,
+		unsigned int *m, unsigned int *s)
+{
+	unsigned int from = 0, to;
+
+	if (!dc || !m) return;
+	for (*m = 0; *m < dc->count; (*m)++) {
+		to = from + mnote_canon_entry_count_values (&dc->entries[*m]);
+		if (to > n) {
+			if (s) *s = n - from;
+			break;
+		}
+		from = to;
+	}
+}
+
 static char *
 exif_mnote_data_canon_get_value (ExifMnoteData *note, unsigned int n, char *val, unsigned int maxlen)
 {
-	ExifMnoteDataCanon *cnote = (ExifMnoteDataCanon *) note;
+	ExifMnoteDataCanon *dc = (ExifMnoteDataCanon *) note;
+	unsigned int m, s;
 
-	if (!note) return NULL;
-	if (cnote->count <= n) return NULL;
-	return mnote_canon_entry_get_value (&cnote->entries[n], val, maxlen);
+	if (!dc) return NULL;
+	exif_mnote_data_canon_get_tags (dc, n, &m, &s);
+	if (m >= dc->count) return NULL;
+	return mnote_canon_entry_get_value (&dc->entries[m], s, val, maxlen);
 }
 
 static void
@@ -74,12 +94,7 @@ exif_mnote_data_canon_set_byte_order (ExifMnoteData *d, ExifByteOrder o)
 {
 	ExifByteOrder o_orig;
 	ExifMnoteDataCanon *n = (ExifMnoteDataCanon *) d;
-	unsigned int i, fs;
-	ExifShort s;
-	ExifLong l;
-	ExifSLong sl;
-	ExifRational r;
-	ExifSRational sr;
+	unsigned int i;
 
 	if (!n) return;
 
@@ -87,55 +102,8 @@ exif_mnote_data_canon_set_byte_order (ExifMnoteData *d, ExifByteOrder o)
 	n->order = o;
 	for (i = 0; i < n->count; i++) {
 		n->entries[i].order = o;
-		fs = exif_format_get_size (n->entries[i].format);
-		switch (n->entries[i].format) {
-		case EXIF_FORMAT_SHORT:
-			for (i = 0; i < n->entries[i].components; i++) {
-				s = exif_get_short (n->entries[i].data + (i*fs),
-						    o_orig);
-				exif_set_short (n->entries[i].data + (i * fs),
-						o, s);
-			}
-			break;
-		case EXIF_FORMAT_LONG:
-			for (i = 0; i < n->entries[i].components; i++) {
-				l = exif_get_long (n->entries[i].data + (i*fs),
-						   o_orig);
-				exif_set_long (n->entries[i].data + (i * fs),
-					       o, l);
-			}
-			break;
-		case EXIF_FORMAT_RATIONAL:
-			for (i = 0; i < n->entries[i].components; i++) {
-				r = exif_get_rational (n->entries[i].data +
-						       (i * fs), o_orig);
-				exif_set_rational (n->entries[i].data +
-					(i * fs), o, r);
-			}
-			break;
-		case EXIF_FORMAT_SLONG:
-			for (i = 0; i < n->entries[i].components; i++) {
-				sl = exif_get_slong (n->entries[i].data +
-						(i * fs), o_orig);
-				exif_set_slong (n->entries[i].data +
-						(i * fs), o, sl);
-			}
-			break;
-		case EXIF_FORMAT_SRATIONAL:
-			for (i = 0; i < n->entries[i].components; i++) {
-				sr = exif_get_srational (n->entries[i].data +
-						(i * fs), o_orig);
-				exif_set_srational (n->entries[i].data +
-						(i * fs), o, sr);
-			}
-			break;
-		case EXIF_FORMAT_UNDEFINED:
-		case EXIF_FORMAT_BYTE:
-		case EXIF_FORMAT_ASCII:
-		default:
-			/* Nothing here. */
-			break;
-		}
+		exif_array_set_byte_order (n->entries[i].format, n->entries[i].data,
+				n->entries[i].components, o_orig, o);
 	}
 }
 
@@ -159,9 +127,8 @@ exif_mnote_data_canon_save (ExifMnoteData *ne,
 	 * of entries.
 	 */
 	*buf_size = 2 + n->count * 12 + 4;
-	*buf = malloc (sizeof (char) * *buf_size);
+	*buf = exif_mem_alloc (ne->mem, sizeof (char) * *buf_size);
 	if (!*buf) return;
-	memset (*buf, 0, sizeof (char) * *buf_size);
 
 	/* Save the number of entries */
 	exif_set_short (*buf, n->order, (ExifShort) n->count);
@@ -178,19 +145,40 @@ exif_mnote_data_canon_save (ExifMnoteData *ne,
 						n->entries[i].components;
 		if (s > 4) {
 			*buf_size += s;
-			*buf = realloc (*buf, sizeof (char) * *buf_size);
+
+			/* Ensure even offsets. Set padding bytes to 0. */
+			if (s & 1) *buf_size += 1;
+			*buf = exif_mem_realloc (ne->mem, *buf,
+						 sizeof (char) * *buf_size);
 			if (!*buf) return;
 			doff = *buf_size - s;
+			if (s & 1) { doff--; *(*buf + *buf_size - 1) = '\0'; }
 			exif_set_long (*buf + o, n->order, n->offset + doff);
 		} else
 			doff = o;
 
-		/* Write the data. Fill unneeded bytes with 0. */
-		memcpy (*buf + doff, n->entries[i].data, s);
+		/*
+		 * Write the data. Fill unneeded bytes with 0. Do not
+		 * crash if data is NULL.
+		 */
+		if (!n->entries[i].data) memset (*buf + doff, 0, s);
+		else memcpy (*buf + doff, n->entries[i].data, s);
 		if (s < 4) memset (*buf + doff + s, 0, (4 - s));
 	}
 }
 
+/* XXX
+ * FIXME: exif_mnote_data_canon_load() may fail and there is no
+ *        semantics to express that.
+ *        See bug #1054323 for details, especially the comment by liblit
+ *        after it has supposedly been fixed:
+ *
+ *        https://sourceforge.net/tracker/?func=detail&aid=1054323&group_id=12272&atid=112272
+ *        Unfortunately, the "return" statements aren't commented at
+ *        all, so it isn't trivial to find out what is a normal
+ *        return, and what is a reaction to an error condition.
+ */
+
 static void
 exif_mnote_data_canon_load (ExifMnoteData *ne,
 	const unsigned char *buf, unsigned int buf_size)
@@ -207,91 +195,106 @@ exif_mnote_data_canon_load (ExifMnoteData *ne,
 
 	/* Parse the entries */
 	for (i = 0; i < c; i++) {
-	    o = 6 + 2 + n->offset + 12 * i;
-	    if (o + 8 > buf_size) return;
+		o = 6 + 2 + n->offset + 12 * i;
+	  if (o + 8 > buf_size) return;
 
-	    n->count = i + 1;
-	    n->entries = realloc (n->entries, sizeof (MnoteCanonEntry) * (i+1));
-	    memset (&n->entries[i], 0, sizeof (MnoteCanonEntry));
-	    n->entries[i].tag        = exif_get_short (buf + o, n->order);
-	    n->entries[i].format     = exif_get_short (buf + o + 2, n->order);
-	    n->entries[i].components = exif_get_long (buf + o + 4, n->order);
-	    n->entries[i].order      = n->order;
+		n->count = i + 1;
+		n->entries = exif_mem_realloc (ne->mem, n->entries,
+				sizeof (MnoteCanonEntry) * (i+1));
+		memset (&n->entries[i], 0, sizeof (MnoteCanonEntry));
+	  n->entries[i].tag        = exif_get_short (buf + o, n->order);
+	  n->entries[i].format     = exif_get_short (buf + o + 2, n->order);
+	  n->entries[i].components = exif_get_long (buf + o + 4, n->order);
+	  n->entries[i].order      = n->order;
 
-	    /*
-	     * Size? If bigger than 4 bytes, the actual data is not
-	     * in the entry but somewhere else (offset).
-	     */
-	    s = exif_format_get_size (n->entries[i].format) *
-		    		      n->entries[i].components;
-	    if (!s) return;
-	    o += 8;
-	    if (s > 4) o = exif_get_long (buf + o, n->order) + 6;
-	    if (o + s > buf_size) return;
-	    
-	    /* Sanity check */
-	    n->entries[i].data = malloc (sizeof (char) * s);
-	    if (!n->entries[i].data) return;
-	    memset (n->entries[i].data, 0, sizeof (char) * s);
-	    n->entries[i].size = s;
-	    memcpy (n->entries[i].data, buf + o, s);
+	  /*
+	   * Size? If bigger than 4 bytes, the actual data is not
+	   * in the entry but somewhere else (offset).
+	   */
+	  s = exif_format_get_size (n->entries[i].format) * n->entries[i].components;
+		if (!s) return;
+		o += 8;
+		if (s > 4) o = exif_get_long (buf + o, n->order) + 6;
+		if (o + s > buf_size) return;
+
+		/* Sanity check */
+		n->entries[i].data = exif_mem_alloc (ne->mem, sizeof (char) * s);
+		if (!n->entries[i].data) return;
+		n->entries[i].size = s;
+		memcpy (n->entries[i].data, buf + o, s);
 	}
 }
 
 static unsigned int
 exif_mnote_data_canon_count (ExifMnoteData *n)
 {
-	return n ? ((ExifMnoteDataCanon *) n)->count : 0;
+	ExifMnoteDataCanon *dc = (ExifMnoteDataCanon *) n;
+	unsigned int i, c;
+
+	for (i = c = 0; dc && (i < dc->count); i++)
+		c += mnote_canon_entry_count_values (&dc->entries[i]);
+	return c;
 }
 
 static unsigned int
-exif_mnote_data_canon_get_id (ExifMnoteData *d, unsigned int n)
+exif_mnote_data_canon_get_id (ExifMnoteData *d, unsigned int i)
 {
-	ExifMnoteDataCanon *note = (ExifMnoteDataCanon *) d;
+	ExifMnoteDataCanon *dc = (ExifMnoteDataCanon *) d;
+	unsigned int m;
 
-	if (!note) return 0;
-	if (note->count <= n) return 0;
-	return note->entries[n].tag;
+	if (!dc) return 0;
+	exif_mnote_data_canon_get_tags (dc, i, &m, NULL);
+	if (m >= dc->count) return 0;
+	return dc->entries[m].tag;
 }
 
 static const char *
 exif_mnote_data_canon_get_name (ExifMnoteData *note, unsigned int i)
 {
-	ExifMnoteDataCanon *cnote = (ExifMnoteDataCanon *) note;
+	ExifMnoteDataCanon *dc = (ExifMnoteDataCanon *) note;
+	unsigned int m, s;
 
-	if (!note) return NULL;
-	if (i >= cnote->count) return NULL;
-	return mnote_canon_tag_get_name (cnote->entries[i].tag);
+	if (!dc) return NULL;
+	exif_mnote_data_canon_get_tags (dc, i, &m, &s);
+	if (m >= dc->count) return NULL;
+	return mnote_canon_tag_get_name_sub (dc->entries[m].tag, s);
 }
 
 static const char *
 exif_mnote_data_canon_get_title (ExifMnoteData *note, unsigned int i)
 {
-	ExifMnoteDataCanon *cnote = (ExifMnoteDataCanon *) note;
+	ExifMnoteDataCanon *dc = (ExifMnoteDataCanon *) note;
+	unsigned int m, s;
 
-	if (!note) return NULL;
-	if (i >= cnote->count) return NULL;
-	return mnote_canon_tag_get_title (cnote->entries[i].tag);
+	if (!dc) return NULL;
+	exif_mnote_data_canon_get_tags (dc, i, &m, &s);
+	if (m >= dc->count) return NULL;
+	return mnote_canon_tag_get_title_sub (dc->entries[m].tag, s);
 }
 
 static const char *
 exif_mnote_data_canon_get_description (ExifMnoteData *note, unsigned int i)
 {
-	ExifMnoteDataCanon *cnote = (ExifMnoteDataCanon *) note;
-	if (!note) return NULL;
-	if (i >= cnote->count) return NULL;
-	return mnote_canon_tag_get_description (cnote->entries[i].tag);
+	ExifMnoteDataCanon *dc = (ExifMnoteDataCanon *) note;
+	unsigned int m;
+
+	if (!dc) return NULL;
+	exif_mnote_data_canon_get_tags (dc, i, &m, NULL);
+	if (m >= dc->count) return NULL;
+	return mnote_canon_tag_get_description (dc->entries[m].tag);
 }
 
 ExifMnoteData *
-exif_mnote_data_canon_new (void)
+exif_mnote_data_canon_new (ExifMem *mem)
 {
 	ExifMnoteData *d;
 
-	d = calloc (1, sizeof (ExifMnoteDataCanon));
+	if (!mem) return NULL;
+
+	d = exif_mem_alloc (mem, sizeof (ExifMnoteDataCanon));
 	if (!d) return NULL;
 
-	exif_mnote_data_construct (d);
+	exif_mnote_data_construct (d, mem);
 
 	/* Set up function pointers */
 	d->methods.free            = exif_mnote_data_canon_free;
