@@ -1,6 +1,8 @@
-/* exif-data.h
+/*! \file exif-data.h
+ * \brief FIXME foo bar blah
  *
- * Copyright  2001 Lutz Mller <lutz@users.sourceforge.net>
+ * \author Lutz Mller <lutz@users.sourceforge.net>
+ * \date 2001-2005
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -26,6 +28,7 @@ extern "C" {
 #endif /* __cplusplus */
 
 #include <libexif/exif-byte-order.h>
+#include <libexif/exif-data-type.h>
 #include <libexif/exif-ifd.h>
 #include <libexif/exif-log.h>
 #include <libexif/exif-tag.h>
@@ -35,6 +38,7 @@ typedef struct _ExifDataPrivate ExifDataPrivate;
 
 #include <libexif/exif-content.h>
 #include <libexif/exif-mnote-data.h>
+#include <libexif/exif-mem.h>
 
 struct _ExifData
 {
@@ -46,7 +50,14 @@ struct _ExifData
 	ExifDataPrivate *priv;
 };
 
-ExifData *exif_data_new   (void);
+ExifData *exif_data_new           (void);
+ExifData *exif_data_new_mem       (ExifMem *);
+
+/*! \brief load exif data from file
+ *  \param[in] path filename including path
+ *  
+ *  Foo bar blah bleh baz.
+ */
 ExifData *exif_data_new_from_file (const char *path);
 ExifData *exif_data_new_from_data (const unsigned char *data,
 				   unsigned int size);
@@ -64,12 +75,26 @@ ExifByteOrder exif_data_get_byte_order  (ExifData *data);
 void          exif_data_set_byte_order  (ExifData *data, ExifByteOrder order);
 
 ExifMnoteData *exif_data_get_mnote_data (ExifData *);
+void           exif_data_fix (ExifData *);
 
 typedef void (* ExifDataForeachContentFunc) (ExifContent *, void *user_data);
 void          exif_data_foreach_content (ExifData *data,
 					 ExifDataForeachContentFunc func,
 					 void *user_data);
 
+typedef enum {
+	EXIF_DATA_OPTION_IGNORE_UNKNOWN_TAGS = 1 << 0,
+	EXIF_DATA_OPTION_FOLLOW_SPECIFICATION = 1 << 1
+} ExifDataOption;
+
+const char *exif_data_option_get_name        (ExifDataOption);
+const char *exif_data_option_get_description (ExifDataOption);
+void        exif_data_set_option             (ExifData *, ExifDataOption);
+void        exif_data_unset_option           (ExifData *, ExifDataOption);
+
+void         exif_data_set_data_type (ExifData *, ExifDataType);
+ExifDataType exif_data_get_data_type (ExifData *);
+
 /* For debugging purposes and error reporting */
 void exif_data_dump (ExifData *data);
 void exif_data_log  (ExifData *data, ExifLog *log);
