@@ -19,8 +19,9 @@
  */
 
 #include <config.h>
-#include "exif-mnote-data.h"
-#include "exif-mnote-data-priv.h"
+
+#include <libexif/exif-mnote-data.h>
+#include <libexif/exif-mnote-data-priv.h>
 
 #include <stdlib.h>
 #include <string.h>
@@ -31,14 +32,17 @@ struct _ExifMnoteDataPriv
 };
 
 void
-exif_mnote_data_construct (ExifMnoteData *d)
+exif_mnote_data_construct (ExifMnoteData *d, ExifMem *mem)
 {
-	if (!d) return;
+	if (!d || !mem) return;
 	if (d->priv) return;
-	d->priv = malloc (sizeof (ExifMnoteDataPriv));
+	d->priv = exif_mem_alloc (mem, sizeof (ExifMnoteDataPriv));
 	if (!d->priv) return;
-	memset (d->priv, 0, sizeof (ExifMnoteDataPriv));
+
 	d->priv->ref_count = 1;
+
+	d->mem = mem;
+	exif_mem_ref (mem);
 }
 
 void
@@ -50,12 +54,17 @@ exif_mnote_data_ref (ExifMnoteData *d)
 static void
 exif_mnote_data_free (ExifMnoteData *d)
 {
+	ExifMem *mem = d ? d->mem : NULL;
+
 	if (!d) return;
 	if (d->priv) {
 		if (d->methods.free) d->methods.free (d);
-		free (d->priv);
+		exif_mem_free (mem, d->priv);
 		d->priv = NULL;
 	}
+	exif_log_unref (d->log);
+	exif_mem_free (mem, d);
+	exif_mem_unref (mem);
 }
 
 void
